<?php

/**
 * Client for VIA-Connect.
 *
 * @category	VIA
 * @package		VIA_Ebay
 * @subpackage	Model
 */
class VIA_Ebay_Model_Client {
	const USE_SANDBOX_CONFIG = 'viaebay/extended/use_sandbox';
	const USE_KEEPALIVE_CONFIG = 'viaebay/extended/use_keepalive';
	const USERNAME_CONFIG = 'viaebay/connection/username';
	const PASSWORD_CONFIG = 'viaebay/connection/password';
	const TOKEN_CONFIG = 'viaebay/connection/token';
	const PROXY_CONFIG = 'viaebay/extended/proxy';
	const PROXY_HOST_CONFIG = 'viaebay/extended/proxy_host';
	const PROXY_PORT_CONFIG = 'viaebay/extended/proxy_port';
	const SANDBOX_URL = 'https://sandboxapi.via.de/';
	const LIVE_URL = 'https://ebayapi.via.de/';
	const LOGIN_URL = 'Authentication_JSON_AppService.axd/Login';
	const SERVICE_URL = 'publicapi/v1/api.svc/';
	
	/**
	 * Use sandbox flag.
	 *
	 * @var boolean
	 */
	protected $useSandbox;
	
	/**
	 * Http Client used to connect to VIA.
	 *
	 * @var Zend_Http_Client
	 */
	protected $httpclient;
	
	/**
	 * Base uri for VIA.
	 * Either sandbox or production.
	 *
	 * @var string
	 */
	protected $baseURI;
	
	/**
	 * Configuration for Http Client.
	 *
	 * @var array
	 */
	protected $config;
	
	/**
	 *
	 * @var VIA_Ebay_Helper_Log
	 */
	protected $log;
	
	/**
	 *
	 * @var VIA_Ebay_Helper_Data
	 */
	protected $helper;
	
	/**
	 * Constructor.
	 *
	 * Load configuration.
	 *
	 * @return void
	 */
	public function __construct() {
		$this->helper = Mage::helper ( 'viaebay/data' );
		$this->log = Mage::helper ( 'viaebay/log' );
		
		$this->config = array (
				'useragent' => 'Magento ' . Mage::getVersion () . ' Extension ' . Mage::helper ( 'viaebay/data' )->getExtensionVersion () 
		);
		
		$this->useSandbox = Mage::getStoreConfig ( self::USE_SANDBOX_CONFIG );
		if ($this->useSandbox) {
			$this->baseURI = self::SANDBOX_URL;
		} else {
			$this->baseURI = self::LIVE_URL;
		}
		
		$this->httpclient = new Zend_Http_Client ();
		$this->httpclient->setHeaders ( array (
				'Content-Type' => 'application/json',
				'Accept' => 'application/json',
				'SubscriptionToken' => trim ( Mage::getStoreConfig ( self::TOKEN_CONFIG ) ),
				'DataServiceVersion' => '1.0',
				'MaxDataServiceVersion' => '2.0',
				'Vendor' => 'Magento ' . Mage::getVersion (),
				'Version' => Mage::helper ( 'viaebay/data' )->getExtensionVersion () 
		) );
		
		if (Mage::getStoreConfig ( self::USE_KEEPALIVE_CONFIG )) {
			$this->log->log ( 'Using http keepalive and persistent tcp', Zend_Log::DEBUG );
			$this->config ['persistent'] = true;
			$this->config ['keepalive'] = true;
		}
		
		if (Mage::getStoreConfigFlag ( self::PROXY_CONFIG )) {
			$host = Mage::getStoreConfig ( self::PROXY_HOST_CONFIG );
			$port = Mage::getStoreConfig ( self::PROXY_PORT_CONFIG );
			
			$this->config ['adapter'] = 'Zend_Http_Client_Adapter_Proxy';
			
			if (strlen ( $host ) > 0) {
				$this->config ['proxy_host'] = $host;
			} else {
				$this->config ['proxy_host'] = '127.0.0.1';
			}
			
			if (strlen ( $port ) > 0) {
				$this->config ['proxy_port'] = $port;
			} else {
				$this->config ['proxy_port'] = 8888;
			}
		}
		
		$this->httpclient->setConfig ( $this->config );
		
		$this->httpclient->setCookieJar ();
		$this->login ();
	}
	
	/**
	 * Login to VIA and store cookie.
	 *
	 * @throws Exception
	 * @return mixed
	 */
	public function login() {
		$username = Mage::getStoreConfig ( self::USERNAME_CONFIG );
		$passwordCrypted = Mage::getStoreConfig ( self::PASSWORD_CONFIG );
		$password = Mage::helper ( 'core' )->decrypt ( $passwordCrypted );
		
		$auth = array (
				'userName' => trim ( $username ),
				'password' => trim ( $password ),
				'createPersistentCookie' => false 
		);
		
		$response = $this->request ( $this->getLoginURI (), null, $auth, Zend_Http_Client::POST );
		if (! $response) {
			throw new Exception ( 'Login not successful' );
		}
		
		return $response;
	}
	
	/**
	 * Load a single entity.
	 *
	 * @param unknown $collection        	
	 * @param unknown $key        	
	 * @param string $filter        	
	 * @param string $expand        	
	 * @return mixed
	 */
	public function loadEntity($collection, $key, $filter = null, $expand = null) {
		$requestParams = array ();
		if ($expand !== null) {
			$requestParams ['$expand'] = $expand;
		}
		if ($filter !== null) {
			$requestParams ['$filter'] = $filter;
		}
		return $this->request ( $this->getEntityUri ( $collection, $key ), $requestParams, null, Zend_Http_Client::GET );
	}
	
	/**
	 * Load entity collection.
	 *
	 * @param unknown $collectionName        	
	 * @param string $filter        	
	 * @param string $expand        	
	 * @return mixed
	 */
	public function loadEntityCollection($collectionName, $filter = null, $expand = null) {
		$requestParams = array ();
		if ($expand !== null) {
			$requestParams ['$expand'] = $expand;
		}
		if ($filter !== null) {
			$requestParams ['$filter'] = $filter;
		}
		return $this->request ( $this->joinURI ( $this->getServiceURI (), $collectionName ), $requestParams, null, Zend_Http_Client::GET );
	}
	
	/**
	 * Load catalogs collection.
	 *
	 * @param string $filter        	
	 * @param string $expand        	
	 * @return mixed
	 */
	public function loadCatalogs($filter = null, $expand = null) {
		return $this->loadEntityCollection ( 'Catalogs', $filter, $expand );
	}
	
	/**
	 * Load products collection.
	 *
	 * @param string $filter        	
	 * @param string $expand        	
	 * @return mixed
	 */
	public function loadProducts($filter = null, $expand = null) {
		return $this->loadEntityCollection ( 'Products', $filter, $expand );
	}
	
	/**
	 * Load product variations collection.
	 *
	 * @param string $filter        	
	 * @param string $expand        	
	 * @return mixed
	 */
	public function loadProductVariations($filter = null, $expand = null) {
		return $this->loadEntityCollection ( 'ProductVariations', $filter, $expand );
	}
	
	/**
	 * Load orders collection.
	 *
	 * @param string $filter        	
	 * @param string $expand        	
	 * @return mixed
	 */
	public function loadOrders($filter = null, $expand = null) {
		return $this->loadEntityCollection ( 'SalesOrders', $filter, $expand );
	}
	
	/**
	 * Call a service operation.
	 *
	 * @param unknown $uri        	
	 * @param array $params        	
	 * @param string $uriIsRelative        	
	 * @return mixed
	 */
	public function call($uri, array $params, $uriIsRelative = true) {
		if ($uriIsRelative) {
			$uri = $this->joinURI ( $this->getServiceURI (), $uri );
		}
		
		return $this->request ( $uri, $params, null, Zend_Http_Client::POST );
	}
	
	/**
	 * Save a single entities to VIA.
	 *
	 * @param unknown $uri        	
	 * @param unknown $data        	
	 * @param string $uriIsRelative        	
	 * @return mixed
	 */
	public function saveObject($uri, $data, $uriIsRelative = true) {
		if ($uriIsRelative) {
			$uri = $this->joinURI ( $this->getServiceURI (), $uri );
		}
		
		return $this->request ( $uri, null, $data, Zend_Http_Client::POST );
	}
	
	/**
	 * Update a single entities by its collection id.
	 *
	 * @param unknown $id        	
	 * @param unknown $data        	
	 * @return mixed
	 */
	public function updateObjectByCollectionId($id, $data) {
		return $this->updateObject ( $this->joinURI ( $this->getServiceURI (), $id ), $data );
	}
	
	/**
	 * Update an entities.
	 *
	 * @param unknown $uri        	
	 * @param unknown $data        	
	 * @return mixed
	 */
	public function updateObject($uri, $data) {
		if (is_array ( $uri )) {
			$uri = $this->resolvUri ( $uri );
		}
		return $this->request ( $uri, null, $data, Zend_Http_Client::MERGE );
	}
	
	/**
	 * Update a single entities field.
	 *
	 * @param unknown $uri        	
	 * @param unknown $field        	
	 * @param unknown $data        	
	 * @return mixed
	 */
	public function updateObjectField($uri, $field, $data) {
		if (! is_array ( $data )) {
			$data = array (
					$field => $data 
			);
		}
		
		return $this->request ( $this->joinURI ( $uri, $field ), null, $data, Zend_Http_Client::MERGE );
	}
	
	/**
	 * Delete given entities.
	 *
	 * @param unknown $uri        	
	 * @param string $key        	
	 * @return mixed
	 */
	public function deleteObject($uri, $key = null) {
		if (is_array ( $uri )) {
			$uri = $this->resolvUri ( $uri );
		} else if ($key != null) {
			$uri = $this->getEntityUri ( $uri, $key );
		}
		return $this->request ( $uri, null, null, Zend_Http_Client::DELETE );
	}
	
	/**
	 * Add a link between two entities.
	 *
	 * @param array $source        	
	 * @param unknown $collection        	
	 * @param array $target        	
	 * @return mixed
	 */
	public function addLink(array $source, $collection, array $target) {
		$uri = $this->joinURI ( $this->resolvUri ( $source ), '$links', $collection );
		return $this->request ( $uri, null, array (
				'uri' => $this->resolvUri ( $target ) 
		), Zend_Http_Client::POST );
	}
	
	/**
	 * Delete a link between two entities.
	 *
	 * @param array $source        	
	 * @param unknown $collection        	
	 * @param array $target        	
	 * @return mixed
	 */
	public function deleteLink(array $source, $collection, array $target) {
		$targetUrl = $this->resolvUri ( $target );
		$targetId = strrchr ( $targetUrl, '(' );
		$uri = $this->joinURI ( $this->resolvUri ( $source ), '$links', $collection . $targetId );
		return $this->request ( $uri, null, null, Zend_Http_Client::DELETE );
	}
	
	/**
	 * Internal request method.
	 * This method calls VIA.
	 *
	 * @param unknown $uri        	
	 * @param unknown $requestParams        	
	 * @param unknown $data        	
	 * @param unknown $method        	
	 * @throws VIA_Ebay_Model_Exception_Client
	 * @return mixed
	 */
	protected function request($uri, $requestParams, $data, $method) {
		if (is_array ( $data ) || is_object ( $data )) {
			$data = Zend_Json::encode ( $data );
		}
		
		$this->httpclient->resetParameters ();
		$this->httpclient->setHeaders ( 'Content-Type', 'application/json' );
		
		if ($requestParams != null) {
			$this->httpclient->setParameterGet ( $requestParams );
		}
		
		if ($data != null) {
			$this->httpclient->setRawData ( $data, 'application/json' );
		}
		
		if (is_array ( $uri )) {
			$uri = $this->resolvUri ( $uri );
		}
		
		$this->httpclient->setUri ( $uri );
		
		$this->log->logToFile ( 'URI: ' . $uri, Zend_Log::DEBUG );
		
		$response = $this->httpclient->request ( $method );
		
		$this->log->logToFile ( "Request\n" . $this->httpclient->getLastRequest (), Zend_Log::DEBUG );
		$this->log->logToFile ( "Response\n" . $this->httpclient->getLastResponse (), Zend_Log::DEBUG );
		
		$responseBody = $response->getBody ();
		
		if (empty ( $responseBody )) {
			return null;
		} else {
			try {
				$decodedResponse = Zend_Json::decode ( $response->getBody () );
			} catch ( Zend_Json_Exception $e ) {
				$this->log->logToFile ( "Cannot decode response body: " + $response->getBody (), Zend_Log::ERR );
				throw $e;
			}
		}
		
		if ($response->isSuccessful ()) {
			if ($response->getHeader ( 'DataServiceVersion' ) == '2.0') {
				// Ignore collection metadata for now
				return $decodedResponse ['d'] ['results'];
			} else {
				return $decodedResponse ['d'];
			}
		} else {
			$message = $this->helper->__ ( "No Error Message" );
			if (array_key_exists ( "error", $decodedResponse ) && array_key_exists ( "message", $decodedResponse ['error'] ) && array_key_exists ( "value", $decodedResponse ['error'] ['message'] )) {
				$message = " " . $decodedResponse ['error'] ['message'] ['value'];
			}
			
			$this->log->logToFile ( $method . " " . $uri . " " . $response->getStatus () . "\n" . $data . "\n" . $response->getBody (), Zend_Log::ERR );
			
			throw new VIA_Ebay_Model_Exception_Client ( $method, $uri, $data, $response->getBody (), $response->getStatus (), $message );
		}
	}
	
	/**
	 * Returns the url using the collection.
	 *
	 * @param unknown $collection        	
	 * @param unknown $key        	
	 * @return Ambigous <NULL, string, mixed>
	 */
	public function getEntityUri($collection, $key) {
		return $this->joinURI ( $this->getServiceURI (), $collection . '(' . $key . 'L)' );
	}
	
	/**
	 * Returns the login uri.
	 *
	 * @return string
	 */
	public function getLoginURI() {
		return $this->getBaseURI () . self::LOGIN_URL;
	}
	
	/**
	 * Returns the service uri.
	 *
	 * @return string
	 */
	public function getServiceURI() {
		return $this->getBaseURI () . self::SERVICE_URL;
	}
	
	/**
	 * Returns the base uri.
	 *
	 * @return string
	 */
	public function getBaseURI() {
		return $this->baseURI;
	}
	
	/**
	 * Resolve uri by entity array.
	 *
	 * @param array $uriObj        	
	 */
	protected function resolvUri(array $uriObj) {
		return $uriObj ['__metadata'] ['uri'];
	}
	
	/**
	 * Join given parameters to a url.
	 * The first parameter is used as base. If the fist parameter is an array the uri is resolved using $this->resolvUri.
	 * If there is a second parameter and it is an array it's values are concatenated to the base.
	 * If the second parameter is not an array all paramaters are concatenated to the base.
	 *
	 * @return NULL|string
	 */
	protected function joinURI() {
		$args = func_get_args ();
		
		if ($args === false) {
			return null;
		}
		
		$uri = array_shift ( $args );
		
		if ($uri == null) {
			return null;
		}
		
		if (count ( $args ) <= 0) {
			return $uri;
		}
		
		if (is_array ( $uri )) {
			$uri = $this->resolvUri ( $uri );
		}
		
		if (is_array ( $args [0] )) {
			$args = $args [0];
		}
		
		foreach ( $args as $value ) {
			$uri = rtrim ( $uri, '/' ) . '/' . ltrim ( $value, '/' );
		}
		
		return $uri;
	}
}